import re, sys, os
try:
	from urllib.parse import unquote, unquote_plus
except ImportError:
	from urlparse import unquote
	from urllib import unquote_plus	
import json
from hashlib import md5
import random

####################################################################################
# Utitlities
####################################################################################
REQUEST = re.compile ('([^ ]+) ([^ ]+)(( HTTP/([0-9.]+))$|$)')
CONNECTION = re.compile ('Connection: (.*)', re.IGNORECASE)

def crack_query (r):
	if type (r) is bytes:
		r = r.decode ("utf8")

	if not r: return {}
	if r[0]=='?': r=r[1:]	
	arg={}
	q = [x.split('=', 1) for x in r.split('&')]
	
	for each in q:
		k = unquote_plus (each[0])
		try: 
			t, k = k.split (":", 1)
		except ValueError:
			t = "str"
			
		if len (each) == 2:		
			v = unquote_plus(each[1])
			if t == "str":
				pass
			elif t == "int":
				v = int (v)
			elif t == "float":
				v = float (v)
			elif t == "list":
				v = v.split (",")
			elif t == "bit":
				v = v.lower () in ("1", "true", "yes")
			elif t == "json":
				v = json.loads (v)
				
		else:
			v = ""			
			if t == "str":
				pass
			elif t == "int":
				v = 0
			elif t == "float":
				v = 0.0
			elif t == "list":
				v = []
			elif t == "bit":
				v = False
			elif t == "json":
				v = {}
				
		if k in arg:
			if type (arg [k]) is not type ([]):
				arg[k] = [arg[k]]
			arg[k].append (v)
			
		else:
			arg[k] = v
			
	return arg

def crack_request (r):
	m = REQUEST.match (r)
	if m.end() == len(r):		
		uri=m.group(2)		
		if m.group(3):
			version = m.group(5)
		else:
			version = None	
		return m.group(1).lower(), uri, version
	else:
		return None, None, None

def join_headers (headers):
	r = []
	for i in range(len(headers)):
		if headers[i][0] in ' \t':	
			r[-1] = r[-1] + headers[i][1:]
		else:
			r.append (headers[i])
	return r

def get_header (head_reg, lines, group=1):
	for line in lines:
		m = head_reg.match (line)
		if m and m.end() == len(line):
			return m.group (group)
	return ''

def get_header_match (head_reg, lines):
	for line in lines:
		m = head_reg.match (line)
		if m and m.end() == len(line):
			return m
	return ''		

def get_extension (path):
	dirsep = path.rfind ('/')
	dotsep = path.rfind ('.')
	if dotsep > dirsep:
		return path[dotsep+1:]
	else:
		return ''

ALNUM = '0123456789abcdefghijklmnopqrstuvwxyz'
def md5uniqid (length = 13):	
	global ALNUM
	_id = ''
	for i in range (0, length):
		_id += random.choice(ALNUM)
	return md5 (_id.encode ("utf8")).hexdigest ()[length:]

def parse_params (header_val):
	d = {}
	header_filed = ""
	last_key = None
	for each in header_val.split (";"):
		if last_key:
			d [last_key] += ";" + each
			if each [-1] == '"':
				last_key = None
			continue
		
		each = each.strip ()
		if not each: continue
		
		try:
			a, b = each.split ("=", 1)
		except ValueError:
			if header_filed:
				header_filed += ":" + each
			else:	
				header_filed = each
			continue
			
		key = a.lower ()
		d [key] = b
		if len (b) >= 2 and  b [0] == '"' and b [-1] != '"':
			last_key = key	
	
	return header_filed, d



"""
$Id: Base.py,v 1.12 2002/04/23 06:04:27 anthonybaxter Exp $

This file is part of the pydns project.
Homepage: http://pydns.sourceforge.net

This code is covered by the standard Python License.

	Base functionality. Request and Response classes, that sort of thing.
"""

import socket, string, types, time
from . import Type,Class,Opcode
import asyncore

class DNSError(Exception): pass

defaults= { 'protocol':'udp', 'port':53, 'opcode':Opcode.QUERY,
			'qtype':Type.A, 'rd':1, 'timing':1, 'timeout': 30, 'addr': None }

defaults['server']=[]

def ParseResolvConf(resolv_path="/etc/resolv.conf"):
	"parses the /etc/resolv.conf file and sets defaults for name servers"
	global defaults
	lines=open(resolv_path).readlines()
	for line in lines:
		line = line.strip()
		if not line or line[0]==';' or line[0]=='#':
			continue
		fields=line.split()
		if len(fields) == 0: 
			continue
		if fields[0]=='domain':
			defaults['domain']=fields[1]
		if fields[0]=='search':
			pass
		if fields[0]=='options':
			pass
		if fields[0]=='sortlist':
			pass
		if fields[0]=='nameserver':
			defaults['server'].append(fields[1])

def DiscoverNameServers():
	import sys
	if sys.platform in ('win32', 'nt'):
		from . import win32dns
		defaults['server'] = win32dns.RegistryResolve()
	else:
		return ParseResolvConf()

class DnsRequest:
	""" high level Request object """
	def __init__(self,*name,**args):
		self.donefunc=None
		self._async=None
		self.defaults = {}
		self.argparse(name,args)
		self.defaults = self.args

	def argparse(self,name,args):
		if not name and 'name' in self.defaults:
			args['name'] = self.defaults['name']
		if type(name) in (str, bytes):
			args['name']=name
		else:
			if len(name) == 1:
				if name[0]:
					args['name']=name[0]
		for i in list(defaults.keys()):
			if i not in args:
				if i in self.defaults:
					args[i]=self.defaults[i]
				else:
					args[i]=defaults[i]
		if type(args['server']) == bytes:
			args['server'] = [args['server']]
		self.args=args

	def socketInit(self,a,b):
		self.s = socket.socket(a,b)

	def processUDPReply(self):
		import time,select
		if self.args['timeout'] > 0:
			r,w,e = select.select([self.s],[],[],self.args['timeout'])
			if not len(r):
				raise DNSError('Timeout')
		self.reply = self.s.recv(1024)
		self.time_finish=time.time()
		self.args['server']=self.ns
		return self.processReply()

	def processTCPReply(self):
		import time
		import Lib
		self.f = self.s.makefile('r')
		header = self.f.read(2)
		if len(header) < 2:
			raise DNSError('EOF')
		count = Lib.unpack16bit(header)
		self.reply = self.f.read(count)
		if len(self.reply) != count:
			raise DNSError('incomplete reply')
		self.time_finish=time.time()
		self.args['server']=self.ns
		return self.processReply()

	def processReply(self):
		import Lib
		self.args['elapsed']=(self.time_finish-self.time_start)*1000
		u = Lib.Munpacker(self.reply)
		r = Lib.DnsResult(u,self.args)
		r.args=self.args
		#self.args=None  # mark this DnsRequest object as used.
		return r
		#### TODO TODO TODO ####
#		if protocol == 'tcp' and qtype == Type.AXFR:
#			while 1:
#				header = f.read(2)
#				if len(header) < 2:
#					print '========== EOF =========='
#					break
#				count = Lib.unpack16bit(header)
#				if not count:
#					print '========== ZERO COUNT =========='
#					break
#				print '========== NEXT =========='
#				reply = f.read(count)
#				if len(reply) != count:
#					print '*** Incomplete reply ***'
#					break
#				u = Lib.Munpacker(reply)
#				Lib.dumpM(u)

	def conn(self):
		self.s.connect((self.ns,self.port))

	def req (self,*name,**args):
		" needs a refactoring "
		import time
		import Lib
		self.argparse(name,args)
		#if not self.args:
		#	raise DNSError,'reinitialize request before reuse'
		protocol = self.args['protocol']
		self.port = self.args['port']
		opcode = self.args['opcode']
		rd = self.args['rd']
		server=self.args['server']
		if type(self.args['qtype']) == bytes:
			try:
				qtype = getattr(Type, self.args['qtype'].upper())
			except AttributeError:
				raise DNSError('unknown query type')
		else:
			qtype=self.args['qtype']
			
		if 'name' not in self.args:
			#print(self.args)
			raise DNSError('nothing to lookup')
			
		qname = self.args['name']
		if qtype == Type.AXFR:
			print('Query type AXFR, protocol forced to TCP')
			protocol = 'tcp'
			
		#print 'QTYPE %d(%s)' % (qtype, Type.typestr(qtype))
		m = Lib.Mpacker()
		# jesus. keywords and default args would be good. TODO.
		m.addHeader(0,
			  0, opcode, 0, 0, rd, 0, 0, 0,
			  1, 0, 0, 0)
		
		m.addQuestion(qname, qtype, Class.IN)
		
		self.request = m.getbuf()
		if protocol == 'udp':
			self.sendUDPRequest(server)
		else:
			self.sendTCPRequest(server)
			
		if self._async:
			return None
			
		else:
			return self.response

	def sendUDPRequest(self, server):
		"refactor me"
		self.response=None
		self.socketInit(socket.AF_INET, socket.SOCK_DGRAM)
		for self.ns in server:
			try:
				# TODO. Handle timeouts &c correctly (RFC)
				#self.s.connect((self.ns, self.port))
				self.conn()
				self.time_start=time.time()
				if not self._async:
					self.s.send(self.request)
					self.response=self.processUDPReply()
			#except socket.error:
			except None:
				continue
			break
		if not self.response:
			if not self._async:
				raise DNSError('no working nameservers found')

	def sendTCPRequest(self, server):
		" do the work of sending a TCP request "
		self.response=None
		for self.ns in server:
			try:
				self.socketInit(socket.AF_INET, socket.SOCK_STREAM)
				self.time_start=time.time()
				self.conn()
				self.s.send(Lib.pack16bit(len(self.request))+self.request)
				self.s.shutdown(1)
				self.response=self.processTCPReply()
			except socket.error:
				continue
			break
		
		if not self.response:
			raise DNSError('no working nameservers found')

#class DnsAsyncRequest(DnsRequest):
class DnsAsyncRequest(DnsRequest,asyncore.dispatcher_with_send):
	" an asynchronous request object. out of date, probably broken "
	def __init__(self,*name,**args):
		DnsRequest.__init__(self, *name, **args)		
		asyncore.dispatcher_with_send.__init__(self)
		# XXX todo
		if 'done' in args and args['done']:
			self.donefunc=args['done']
		else:
			self.donefunc=self.showResult
		#self.realinit(name,args) # XXX todo
		self._async=1
	
	def conn(self):
		import time
		self.connect((self.ns,self.port))
		self.time_start=time.time()
		if 'start' in self.args and self.args['start']:
			asyncore.dispatcher.go(self)
	
	def socketInit (self,a,b):		
		self.create_socket(a,b)
		self.s=self
	
	def handle_read(self):
		if self.args['protocol'] == 'udp':
			self.response=self.processUDPReply()
			if self.donefunc:
				self.donefunc(*(self,))
	
	def handle_connect(self):
		self.send(self.request)
	
	def handle_write(self):
		pass
	
	def showResult(self,*s):
		self.response.show()

#
# $Log: Base.py,v $
# Revision 1.12  2002/04/23 06:04:27  anthonybaxter
# attempt to refactor the DNSRequest.req method a little. after doing a bit
# of this, I've decided to bite the bullet and just rewrite the puppy. will
# be checkin in some design notes, then unit tests and then writing the sod.
#
# Revision 1.11  2002/03/19 13:05:02  anthonybaxter
# converted to class based exceptions (there goes the python1.4 compatibility :)
#
# removed a quite gross use of 'eval()'.
#
# Revision 1.10  2002/03/19 12:41:33  anthonybaxter
# tabnannied and reindented everything. 4 space indent, no tabs.
# yay.
#
# Revision 1.9  2002/03/19 12:26:13  anthonybaxter
# death to leading tabs.
#
# Revision 1.8  2002/03/19 10:30:33  anthonybaxter
# first round of major bits and pieces. The major stuff here (summarised
# from my local, off-net CVS server :/ this will cause some oddities with
# the
#
# tests/testPackers.py:
#   a large slab of unit tests for the packer and unpacker code in DNS.Lib
#
# DNS/Lib.py:
#   placeholder for addSRV.
#   added 'klass' to addA, make it the same as the other A* records.
#   made addTXT check for being passed a string, turn it into a length 1 list.
#   explicitly check for adding a string of length > 255 (prohibited).
#   a bunch of cleanups from a first pass with pychecker
#   new code for pack/unpack. the bitwise stuff uses struct, for a smallish
#	 (disappointly small, actually) improvement, while addr2bin is much
#	 much faster now.
#
# DNS/Base.py:
#   added DiscoverNameServers. This automatically does the right thing
#	 on unix/ win32. No idea how MacOS handles this.  *sigh*
#	 Incompatible change: Don't use ParseResolvConf on non-unix, use this
#	 function, instead!
#   a bunch of cleanups from a first pass with pychecker
#
# Revision 1.5  2001/08/09 09:22:28  anthonybaxter
# added what I hope is win32 resolver lookup support. I'll need to try
# and figure out how to get the CVS checkout onto my windows machine to
# make sure it works (wow, doing something other than games on the
# windows machine :)
#
# Code from Wolfgang.Strobl@gmd.de
# win32dns.py from
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66260
#
# Really, ParseResolvConf() should be renamed "FindNameServers" or
# some such.
#
# Revision 1.4  2001/08/09 09:08:55  anthonybaxter
# added identifying header to top of each file
#
# Revision 1.3  2001/07/19 07:20:12  anthony
# Handle blank resolv.conf lines.
# Patch from Bastian Kleineidam
#
# Revision 1.2  2001/07/19 06:57:07  anthony
# cvs keywords added
#
#



from ...athreads.fifo import await_fifo, await_ts_fifo

class http2_producer_fifo (await_fifo):
	# asyncore await_fifo replacement
	# for resorting, removeing by http2 priority, cacnel and reset features
	# also can handle producers has 'ready' method
	# this class can be used only at single thread environment

	def remove (self, stream_id):
		if self.l:
			for i in range (len (self.l)):
				try:
					producer_stream_id = self.l [0].stream_id
				except AttributeError:
					pass
				else:
					if producer_stream_id	== stream_id:
						self.l.popleft ()							
				self.l.rotate (1)				
			
	def _insert_into (self, index, item):		
		if index == 0:
			self.l.appendleft (item)
		elif index == -1:
			self.l.append (item)	
		else:
			r = len (self.l) - index
			self.l.rotate (r)
			self.l.append (item)
			self.l.rotate (index + 1)
					
	def insert (self, index, item):
		if item is None:
			self.has_None = True
			return
		
		if self.has_None and index != 0:
			return # deny adding			
		if not self.l:
			return self.l.append (item)
				
		if index == 0:
			try:
				readyfunc = getattr (item, "ready")
			except AttributeError:
				pass
			else:		
				if not readyfunc ():
					return self.l.append (item)
						
			return self.l.appendleft (item)
		
		# insert by priority
		try:
			d1 = item.depends_on
			w1 = item.weight			
		except AttributeError:
			return self.l.append (item)
		
		if d1 == 0:
			return self.l.append (item)
		
		i = 0
		found_parent = 0			
		for each in self.l:
			try:
				s2 = each.stream_id					
			except AttributeError:
				pass
			else:
				if found_parent:
					d2 = each.depends_on
					w2 = each.weight					
					if d1 == d2 and w2 < w1:
						return self._insert_into (i, item)
				elif d1 == s2:
					found_parent = 1
			i += 1
			
		self.l.append (item)


class http2_producer_ts_fifo (http2_producer_fifo, await_ts_fifo):	
	# multin threads safe version	
	
	def remove (self, stream_id):
		with self._lock:			
			http2_producer_fifo.remove (self, stream_id)
					
	def insert (self, index, item):		
		with self._lock:			
			http2_producer_fifo.insert (self, index, item)
	


import re
import sys
from rs4 import strutil
HAS_SKILLSET = True
try:
	import html5lib	
	import lxml.etree
	import lxml.html	
except ImportError:
	HAS_SKILLSET = False
import warnings
warnings.filterwarnings("ignore")

def remove_control_characters (html):
	def str_to_int(s, default, base=10):
		if int(s, base) < 0x10000:
			if strutil.PY_MAJOR_VERSION == 2:
				return unichr(int(s, base))
			else:	
				return chr(int(s, base))				
		return default

	html = re.sub(r"&#(\d+);?", lambda c: str_to_int(c.group(1), c.group(0)), html)
	html = re.sub(r"&#[xX]([0-9a-fA-F]+);?", lambda c: str_to_int(c.group(1), c.group(0), base=16), html)
	html = re.sub(r"[\x00-\x08\x0b\x0e-\x1f\x7f]", "", html)		
	return html
	
def remove_non_asc (html):	
	html = re.sub(br"&#(\d+);?", "", html)
	html = re.sub(br"&#[xX]([0-9a-fA-F]+);?", "", html)
	html = re.sub(br"[\x00-\x08\x80-\xff]", "", html)	
	return html
	
RX_CAHRSET = re.compile (br"[\s;]+charset\s*=\s*['\"]?([-a-z0-9]+)", re.M) #"
RX_META = re.compile (br"<meta\s+.+?>", re.I|re.M)

def get_charset (html):	
	encoding = None
	pos = 0	
	while 1:	
		match = RX_META.search (html, pos)
		if not match: break
		#print (match.group ())	
		charset = RX_CAHRSET.findall (match.group ().lower ())
		if charset:			
			encoding = charset [0].decode ("utf8")			
			#print (encoding)
			break
		pos = match.end ()		
	return encoding

def to_str (body, encoding = None):
	def try_generic_encoding (html):
		try:
			return html.decode ("utf8")
		except UnicodeDecodeError:	
			try:
				return html.decode ("iso8859-1")
			except UnicodeDecodeError:
				return remove_non_asc (html).decode ("utf8")
	
	if type (body) is bytes:	
		if encoding:
			try:
				body = body.decode (encoding)
			except (UnicodeDecodeError, LookupError):
				inline_encoding = get_charset (body)
				if inline_encoding and encoding != inline_encoding:				
					try:
						body = body.decode (inline_encoding)
					except (UnicodeDecodeError, LookupError):
						body = try_generic_encoding (body)			
				else:
					body = try_generic_encoding (body)
		else:
			body = try_generic_encoding (body)
	
	return remove_control_characters (body)
	
def html (html, baseurl, encoding = None):
	# html5lib rebuilds possibly mal-formed html	
	try:		
		if type (html) is str:
			return lxml.html.fromstring (lxml.etree.tostring (html5lib.parse (html, treebuilder="lxml")), baseurl)
		else:	
			return lxml.html.fromstring (lxml.etree.tostring (html5lib.parse (html, likely_encoding = encoding, treebuilder="lxml")), baseurl)
	except ValueError:
		return lxml.html.fromstring (lxml.etree.tostring (html5lib.parse (to_str (html, encoding), treebuilder="lxml")), baseurl)

def etree (html, encoding = None):
	try:
		return html5lib.parse (html, encoding = encoding, treebuilder="lxml")	
	except ValueError:	
		return html5lib.parse (to_str (html, encoding), treebuilder="lxml")	


if __name__ == "__main__":	
	from urllib.request import urlopen
	from contextlib import closing
	
	with closing(urlopen("http://www.drugandalcoholrehabhouston.com")) as f:
		build (f.read ())

